@@ -33,8 +33,14 @@
 
 ;;; Find Blob
 
-(defvar magit-find-file-hook nil)
-(add-hook 'magit-find-file-hook #'magit-blob-mode)
+(define-obsolete-variable-alias 'magit-find-file-hook
+  'magit-find-blob-hook "Magit 4.6.0")
+
+(define-obsolete-variable-alias 'magit-find-index-hook
+  'magit-find-blob-hook "Magit 4.6.0")
+
+(defvar magit-find-blob-hook nil)
+(add-hook 'magit-find-blob-hook #'magit-blob-mode)
 
 ;;;###autoload
 (defun magit-find-file (rev file)
@@ -44,7 +50,7 @@ already exists.  If prior to calling this command the current
 buffer and/or cursor position is about the same file, then go
 to the line and column corresponding to that location."
   (interactive (magit-find-file-read-args "Find file"))
-  (magit-find-file--internal rev file #'pop-to-buffer-same-window))
+  (pop-to-buffer-same-window (magit-find-file-noselect rev file)))
 
 ;;;###autoload
 (defun magit-find-file-other-window (rev file)
@@ -54,7 +60,7 @@ already exists.  If prior to calling this command the current
 buffer and/or cursor position is about the same file, then go to
 the line and column corresponding to that location."
   (interactive (magit-find-file-read-args "Find file in other window"))
-  (magit-find-file--internal rev file #'switch-to-buffer-other-window))
+  (switch-to-buffer-other-window (magit-find-file-noselect rev file)))
 
 ;;;###autoload
 (defun magit-find-file-other-frame (rev file)
@@ -64,82 +70,63 @@ already exists.  If prior to calling this command the current
 buffer and/or cursor position is about the same file, then go to
 the line and column corresponding to that location."
   (interactive (magit-find-file-read-args "Find file in other frame"))
-  (magit-find-file--internal rev file #'switch-to-buffer-other-frame))
+  (switch-to-buffer-other-frame (magit-find-file-noselect rev file)))
 
 (defun magit-find-file-read-args (prompt)
-  (let ((pseudo-revs '("{worktree}" "{index}")))
-    (let ((rev (magit-completing-read "Find file from revision"
-                                      (append pseudo-revs
-                                              (magit-list-refnames nil t))
-                                      nil 'any nil 'magit-revision-history
-                                      (or (magit-branch-or-commit-at-point)
-                                          (magit-get-current-branch)))))
-      (list rev
-            (magit-read-file-from-rev (if (member rev pseudo-revs) "HEAD" rev)
-                                      prompt)))))
+  (let* ((pseudo-revs '("{worktree}" "{index}"))
+         (rev (magit-completing-read "Find file from revision"
+                                     (append pseudo-revs
+                                             (magit-list-refnames nil t))
+                                     nil 'any nil 'magit-revision-history
+                                     (or (magit-branch-or-commit-at-point)
+                                         (magit-get-current-branch)))))
+    (list rev
+          (magit-read-file-from-rev (if (member rev pseudo-revs) "HEAD" rev)
+                                    prompt))))
 
-(defun magit-find-file--internal (rev file fn)
-  (let ((buf (magit-find-file-noselect rev file))
-        line col)
-    (when-let ((visited-file (magit-file-relative-name)))
-      (setq line (line-number-at-pos))
-      (setq col (current-column))
-      (cond
-        ((not (equal visited-file file)))
-        ((equal magit-buffer-revision rev))
-        ((equal rev "{worktree}")
-         (setq line (magit-diff-visit--offset file magit-buffer-revision line)))
-        ((equal rev "{index}")
-         (setq line (magit-diff-visit--offset file nil line)))
-        (magit-buffer-revision
-         (setq line (magit-diff-visit--offset
-                     file (concat magit-buffer-revision ".." rev) line)))
-        ((setq line (magit-diff-visit--offset file (list "-R" rev) line)))))
-    (funcall fn buf)
-    (when line
-      (with-current-buffer buf
-        (widen)
-        (goto-char (point-min))
-        (forward-line (1- line))
-        (move-to-column col)))
-    buf))
-
-(defun magit-find-file-noselect (rev file &optional revert)
+(defun magit-find-file-noselect (rev file &optional no-restore-position)
   "Read FILE from REV into a buffer and return the buffer.
-REV is a revision or one of \"{worktree}\" or \"{index}\".  FILE must
-be relative to the top directory of the repository.  Non-nil REVERT
-means to revert the buffer.  If `ask-revert', then only after asking.
-A non-nil value for REVERT is ignored if REV is \"{worktree}\"."
-  (cond-let*
-    ((equal rev "{worktree}")
-     (let ((revert-without-query
-            (if (and$ (find-buffer-visiting file)
-                      (buffer-local-value 'auto-revert-mode $))
-                (cons "." revert-without-query)
-              revert-without-query)))
-       (find-file-noselect file)))
-    ([topdir (magit-toplevel)]
-     [file   (expand-file-name file topdir)]
-     [defdir (file-name-directory file)]
-     (setq rev (magit--abbrev-if-hash rev))
-     (with-current-buffer (magit-get-revision-buffer-create
-                           rev
-                           (file-relative-name file topdir))
-       (when (or (not magit-buffer-file-name)
-                 (if (eq revert 'ask-revert)
-                     (y-or-n-p (format "%s already exists; revert it? "
-                                       (buffer-name))))
-                 revert)
-         (setq magit-buffer-revision rev)
-         (setq magit-buffer-file-name file)
-         (setq default-directory (if (file-exists-p defdir) defdir topdir))
-         (setq-local revert-buffer-function #'magit-revert-rev-file-buffer)
-         (revert-buffer t t)
-         (run-hooks (if (equal rev "{index}")
-                        'magit-find-index-hook
-                      'magit-find-file-hook)))
-       (current-buffer)))
-    ((error "%s isn't inside a Git repository" file))))
+REV is a revision or one of \"{worktree}\" or \"{index}\"."
+  (let* ((rev (pcase rev
+                ('nil "{worktree}")
+                ((and "{index}"
+                      (guard (length> (magit--file-index-stages file) 1)))
+                 "{worktree}")
+                (rev rev)))
+         (topdir (magit-toplevel))
+         (file (expand-file-name file topdir))
+         (file-relative (file-relative-name file topdir))
+         (buffer
+          (cond-let
+            ((equal rev "{worktree}")
+             (let ((revert-without-query
+                    (if (and$ (find-buffer-visiting file)
+                              (buffer-local-value 'auto-revert-mode $))
+                        (cons "." revert-without-query)
+                      revert-without-query)))
+               (find-file-noselect file)))
+            ((not topdir)
+             (error "%s is not inside a Git repository" file))
+            ([defdir (file-name-directory file)]
+             [rev (magit--abbrev-if-hash rev)]
+             (unless (file-in-directory-p file topdir)
+               (error "%s is not inside Git repository %s" file topdir))
+             (with-current-buffer
+                 (magit-get-revision-buffer-create rev file-relative)
+               (setq magit-buffer-revision rev)
+               (setq magit-buffer-file-name file)
+               (setq default-directory
+                     (if (file-exists-p defdir) defdir topdir))
+               (setq-local revert-buffer-function #'magit--revert-blob-buffer)
+               (magit--refresh-blob-buffer)
+               (current-buffer)))
+            ((error "Unexpected error")))))
+    (when (and (not no-restore-position)
+               (equal (magit-file-relative-name) file-relative))
+      (let ((pos (magit-find-file--position)))
+        (with-current-buffer buffer
+          (apply #'magit-find-file--restore-position pos))))
+    buffer))
 
 (defun magit-get-revision-buffer-create (rev file)
   (magit-get-revision-buffer rev file t))
@@ -148,41 +135,74 @@ A non-nil value for REVERT is ignored if REV is \"{worktree}\"."
   (funcall (if create #'get-buffer-create #'get-buffer)
            (format "%s.~%s~" file (subst-char-in-string ?/ ?_ rev))))
 
-(defun magit-revert-rev-file-buffer (_ignore-auto noconfirm)
-  (when (or noconfirm
-            (and (not (buffer-modified-p))
-                 (catch 'found
-                   (dolist (regexp revert-without-query)
-                     (when (string-match regexp magit-buffer-file-name)
-                       (throw 'found t)))))
-            (yes-or-no-p (format "Revert buffer from Git %s? "
-                                 (if (equal magit-buffer-revision "{index}")
-                                     "index"
-                                   (concat "revision " magit-buffer-revision)))))
-    (let* ((inhibit-read-only t)
-           (default-directory (magit-toplevel))
-           (file (file-relative-name magit-buffer-file-name))
-           (coding-system-for-read (or coding-system-for-read 'undecided)))
-      (erase-buffer)
-      (magit-git-insert "cat-file" "-p"
-                        (if (equal magit-buffer-revision "{index}")
-                            (concat ":" file)
-                          (concat magit-buffer-revision ":" file)))
-      (setq buffer-file-coding-system last-coding-system-used))
-    (let ((buffer-file-name magit-buffer-file-name)
-          (after-change-major-mode-hook
-           (seq-difference after-change-major-mode-hook
-                           '(global-diff-hl-mode-enable-in-buffer ; Emacs >= 30
-                             global-diff-hl-mode-enable-in-buffers ; Emacs < 30
-                             eglot--maybe-activate-editing-mode)
-                           #'eq)))
-      ;; We want `normal-mode' to respect nil `enable-local-variables'.
-      ;; The FIND-FILE argument wasn't designed for our use case, so we
-      ;; have to use this strange invocation to achieve that.
-      (normal-mode (not enable-local-variables)))
+(defun magit--revert-blob-buffer (_ignore-auto _noconfirm)
+  (let ((pos (magit-find-file--position)))
+    (magit--refresh-blob-buffer t)
+    (apply #'magit-find-file--restore-position pos)))
+
+(defun magit--refresh-blob-buffer (&optional force)
+  (let ((old-blob-oid magit-buffer-blob-oid))
+    (setq magit-buffer-revision-oid
+          (magit-commit-oid magit-buffer-revision t))
+    (setq magit-buffer-blob-oid
+          (magit-blob-oid magit-buffer-revision magit-buffer-file-name))
+    (when (or force (not (equal old-blob-oid magit-buffer-blob-oid)))
+      (let ((inhibit-read-only t))
+        (erase-buffer)
+        (save-excursion
+          (magit--insert-blob-contents magit-buffer-revision
+                                       (magit-file-relative-name))))))
+  (magit--blob-normal-mode))
+
+(defun magit--blob-normal-mode ()
+  (let ((buffer-file-name magit-buffer-file-name)
+        (after-change-major-mode-hook
+         ;; Inhibit diff-hl and eglot; see bb8a65269d and 234a787b8c.
+         (seq-difference after-change-major-mode-hook
+                         '(global-diff-hl-mode-enable-in-buffer ; Emacs >= 30
+                           global-diff-hl-mode-enable-in-buffers ; Emacs < 30
+                           eglot--maybe-activate-editing-mode)
+                         #'eq)))
+    ;; We want `normal-mode' to respect nil `enable-local-variables'.
+    ;; The FIND-FILE argument wasn't designed for our use case,
+    ;; so we have to use this strange invocation to achieve that.
+    (normal-mode (not enable-local-variables))
     (setq buffer-read-only t)
     (set-buffer-modified-p nil)
-    (goto-char (point-min))))
+    (run-hooks 'magit-find-blob-hook)))
+
+(defun magit-find-file--position ()
+  (list (or magit-buffer-revision-oid magit-buffer-revision "{worktree}")
+        magit-buffer-blob-oid
+        (line-number-at-pos)
+        (current-column)))
+
+(defun magit-find-file--restore-position (before blob line col)
+  (let ((file (magit-file-relative-name))
+        (rev (or magit-buffer-revision-oid magit-buffer-revision "{worktree}")))
+    (goto-char (point-min))
+    (forward-line
+     (1-
+      (pcase (list before rev)
+        ((guard (equal magit-buffer-blob-oid blob)) line)
+        ('("{worktree}" "{worktree}") line)
+        ('("{worktree}" "{index}")
+         (magit-diff-visit--offset line file "-R"))
+        (`("{worktree}" ,_)
+         (magit-diff-visit--offset line file "-R" rev))
+        ('("{index}" "{worktree}")
+         (magit-diff-visit--offset line file))
+        ('("{index}" "{index}")
+         (magit-diff-visit--offset line (list blob magit-buffer-blob-oid file)))
+        (`("{index}" ,_)
+         (magit-diff-visit--offset line file "-R" "--cached"))
+        (`(,_ "{worktree}")
+         (magit-diff-visit--offset line file before))
+        (`(,_ "{index}")
+         (magit-diff-visit--offset line file "--cached"))
+        (_
+         (magit-diff-visit--offset line file before rev)))))
+    (move-to-column col)))
 
 (define-advice lsp (:around (fn &rest args) magit-find-file)
   "Do nothing when visiting blob using `magit-find-file' and similar.
@@ -190,15 +210,7 @@ See also https://github.com/doomemacs/doomemacs/pull/6309."
   (unless magit-buffer-revision
     (apply fn args)))
 
-;;; Find Index
-
-(defvar magit-find-index-hook nil)
-(add-hook 'magit-find-index-hook #'magit-blob-mode)
-
-(defun magit-find-file-index-noselect (file &optional revert)
-  "Read FILE from the index into a buffer and return the buffer.
-FILE must to be relative to the top directory of the repository."
-  (magit-find-file-noselect "{index}" file (or revert 'ask-revert)))
+;;; Update Index
 
 (defun magit-update-index ()
   "Update the index with the contents of the current buffer.
@@ -346,6 +358,7 @@ to `magit-dispatch'."
 
 (defvar-keymap magit-blob-mode-map
   :doc "Keymap for `magit-blob-mode'."
+  "g" #'revert-buffer
   "p" #'magit-blob-previous
   "n" #'magit-blob-next
   "b" #'magit-blame-addition
@@ -413,7 +426,7 @@ When visiting a blob or the version from the index, then go to
 the same location in the respective file in the working tree."
   (interactive)
   (if-let ((file (magit-file-relative-name)))
-      (magit-find-file--internal "{worktree}" file #'pop-to-buffer-same-window)
+      (pop-to-buffer-same-window (magit-find-file-noselect "{worktree}" file))
     (user-error "Not visiting a blob")))
 
 (defun magit-blob-visit (rev file)
@@ -625,6 +638,17 @@ If DEFAULT is non-nil, use this as the default value instead of
 (define-obsolete-function-alias 'magit-find-file-noselect-1
   'magit-find-file-noselect "Magit 4.4.0")
 
+(defun magit-find-file--internal (rev file display)
+  (declare (obsolete magit-find-file-noselect "Magit 4.6.0"))
+  (let ((buf (magit-find-file-noselect rev file)))
+    (funcall display buf)
+    buf))
+
+(defun magit-find-file-index-noselect (file)
+  "Read FILE from the index into a buffer and return the buffer."
+  (declare (obsolete magit-find-file-noselect "Magit 4.6.0"))
+  (magit-find-file-noselect "{index}" file t))
+
 (provide 'magit-files)
 ;; Local Variables:
 ;; read-symbol-shorthands: (
